@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: UTF-8 -*-
 #
-# Copyright 2015-2022, Vincenzo Arcidiacono;
+# Copyright 2015-2024, Vincenzo Arcidiacono;
 # Licensed under the EUPL (the 'Licence');
 # You may not use this work except in compliance with the Licence.
 # You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
@@ -41,7 +41,6 @@ import functools
 import itertools
 import threading
 import collections
-import pkg_resources
 import os.path as osp
 import urllib.parse as urlparse
 
@@ -73,7 +72,7 @@ PLOT_EXECUTORS = ExecutorFactory({
     'sync': _sync_executor,
     'async': _async_executor
 })
-
+pkg_dir = osp.abspath(osp.dirname(__file__))
 atexit_register(PLOT_EXECUTORS.shutdown_executor, wait=False)
 
 
@@ -116,7 +115,9 @@ def autoplot_callback(res):
 
 
 def jinja2_format(source, context=None, **kw):
-    return Environment(**kw).from_string(source).render(context or {})
+    return Environment(**kw, autoescape=True).from_string(source).render(
+        context or {}
+    )
 
 
 def valid_filename(item, filenames, ext=None):
@@ -148,18 +149,33 @@ def update_filenames(node, filenames):
 
 
 def site_view(
-        app, context, generated_files, rendered, rules, root, filepath=None,
-        viz=False, executor='async'):
+        context, rendered, rules, root, filepath=None, viz=False,
+        executor='async'):
+    from flask import current_app
+    app = current_app
     static_folder = app.static_folder
     filepath = filepath or root
-    if not osp.isfile(osp.join(static_folder, filepath)):
-        if filepath not in rules:
+    render_ctx = {
+        'alive_url': 'alive' in app.view_functions and app.url_for('alive')
+    }
+    render_ctx.update(app.jinja_env.globals)
+    fpath = osp.join(static_folder, filepath)
+    if not osp.isfile(fpath):
+        try:
+            node = rules[filepath]
+        except KeyError:
             from flask import abort
             return abort(404)
-        node = rules[filepath]
-        generated_files.extend((v.result() for v in cached_view(
-            node, static_folder, context, rendered, viz, executor
-        ).values()))
+        rend, expected = cached_view(
+            node, static_folder, context, rendered, viz, executor, **render_ctx
+        )
+        for k, v in rend.items():
+            fpath_rendered = v.result()
+            fpath_expected = expected[k]
+            if fpath_rendered != fpath_expected:
+                os.makedirs(osp.dirname(fpath_expected), exist_ok=True)
+                shutil.copy(fpath_rendered, fpath_expected)
+
     return app.send_static_file(filepath)
 
 
@@ -199,10 +215,7 @@ class SiteNode:
 
     @property
     def name(self):
-        try:
-            return parent_func(self.item).__name__
-        except AttributeError:
-            return self.node_id
+        return self.node_id
 
     @property
     def view_id(self):
@@ -227,7 +240,9 @@ class SiteNode:
         code = render_output(self.item, self.pprint.pformat)
         formatter = HtmlFormatter(noclasses=True)
         formatter.style.background_color = 'transparent'
-        return _format_output(highlight(code, Python3Lexer(), formatter))
+        return _format_output(
+            highlight(code, Python3Lexer(), formatter), **kwargs
+        )
 
     def view(self, filepath, *args, **kwargs):
         fpath = uncpath(filepath)
@@ -265,7 +280,7 @@ class FolderNode:
         'info': {
             START: {
                 'shape': 'egg', 'label': 'start', 'fillcolor': '#39bf7f',
-                'color': '#39bf7f'
+                'color': '#39bf7f', 'ordering': 'out'
             },
             SELF: {
                 'shape': 'egg', 'label': 'self', 'fillcolor': '#C1A4FE',
@@ -277,7 +292,7 @@ class FolderNode:
             },
             END: {
                 'shape': 'egg', 'label': 'end', 'fontcolor': '#FFFFFF',
-                'fillcolor': '#084368', 'color': '#084368'
+                'fillcolor': '#084368', 'color': '#084368', 'ordering': 'in'
             },
             EMPTY: {
                 'shape': 'egg', 'label': 'empty', 'fillcolor': '#FFFFFF'
@@ -292,13 +307,16 @@ class FolderNode:
                     'fillcolor': '#73c4fa', 'color': '#73c4fa'
                 },
                 'function': {
-                    'shape': 'box', 'fillcolor': '#eed867', 'color': '#eed867'
+                    'shape': 'box', 'fillcolor': '#eed867', 'color': '#eed867',
+                    'ordering': 'in'
                 },
                 'function-dispatcher': {
-                    'shape': 'note', 'fillcolor': '#eed867', 'color': '#eed867'
+                    'shape': 'note', 'fillcolor': '#eed867', 'color': '#eed867',
+                    'ordering': 'in'
                 },
                 'run_model': {
-                    'shape': 'note', 'fillcolor': '#eed867', 'color': '#eed867'
+                    'shape': 'note', 'fillcolor': '#eed867', 'color': '#eed867',
+                    'ordering': 'in'
                 },
                 'subdispatch': {
                     'shape': 'note', 'style': 'filled', 'fillcolor': '#ffc490',
@@ -306,19 +324,19 @@ class FolderNode:
                 },
                 'mapdispatch': {
                     'shape': 'note', 'style': 'filled', 'fillcolor': '#f4bd6a',
-                    'color': '#f4bd6a'
+                    'color': '#f4bd6a', 'ordering': 'in'
                 },
                 'subdispatchfunction': {
                     'shape': 'note', 'style': 'filled', 'fillcolor': '#f9d951',
-                    'color': '#f9d951'
+                    'color': '#f9d951', 'ordering': 'in'
                 },
                 'subdispatchpipe': {
                     'shape': 'note', 'style': 'filled', 'fillcolor': '#f1cd5d',
-                    'color': '#f1cd5d'
+                    'color': '#f1cd5d', 'ordering': 'in'
                 },
                 'dispatchpipe': {
                     'shape': 'note', 'style': 'filled', 'fillcolor': '#e8c268',
-                    'color': '#e8c268'
+                    'color': '#e8c268', 'ordering': 'in'
                 },
                 'dispatcher': {
                     'shape': 'note', 'style': 'filled', 'fillcolor': '#c6c6c6',
@@ -656,7 +674,7 @@ class FolderNode:
                         tr.add(**v)
                     else:
                         j = render_output(j, pformat)
-                        s = jinja2_format(j, cnt)
+                        s = html.unescape(jinja2_format(j, cnt))
                         if s.startswith('_Td('):
                             tr += eval(s)
                         else:  # It is not a valid jinja2 format.
@@ -689,21 +707,23 @@ class FolderNode:
 
 
 def _format_output(obj, **kwargs):
-    pkg_dir = pkg_resources.resource_filename(__name__, '')
     fpath = osp.join(pkg_dir, 'templates', 'render.html')
     with open(fpath) as template:
         return jinja2_format(
-            template.read(), combine_dicts({'obj': obj}, kwargs),
+            template.read(), combine_dicts(kwargs, base={'obj': obj}),
             loader=PackageLoader(__name__)
         )
 
 
 def _format_kw_digraph(*dicts, base=None):
     kw = combine_nested_dicts(*dicts, base=base)
-    if 'body' in kw:
-        kw['body'] = [
-            '\t%s = %s\n' % (k, v) for k, v in sorted(kw['body'].items())
-        ]
+    body = kw.pop('body', None)
+    if 'raw_body' in kw:
+        get_nested_dicts(kw, 'body', default=list).extend(kw.pop('raw_body'))
+    if body:
+        get_nested_dicts(kw, 'body', default=list).extend(
+            '\t%s = %s\n' % (k, v) for k, v in sorted(body.items())
+        )
     return kw
 
 
@@ -711,7 +731,9 @@ class SiteFolder:
     counter = SiteNode.counter
     digraph = {
         'node_attr': {'style': 'filled'},
-        'graph_attr': {'bgcolor': 'transparent'},
+        'graph_attr': {
+            'bgcolor': 'transparent', 'nslimit': '1', 'nslimit1': '1'
+        },
         'edge_attr': {},
         'body': {'splines': 'ortho', 'style': 'filled'},
         'format': 'svg'
@@ -811,8 +833,12 @@ class SiteFolder:
 
         if any(o in it for o in self.outputs) and END not in gnode:
             it[END] = {'index': (END,)}
-
-        for k, a in sorted(it.items()):
+        it = sorted(
+            it.items(),
+            key=lambda x: (
+                x[0] is END, nodes.get(x[0], {}).get('index', ()), x[0])
+        )
+        for k, a in it:
             attr = combine_dicts(nodes.get(k, {}), a)
             if k in errors:
                 attr['error'] = errors[k]
@@ -890,7 +916,7 @@ class SiteFolder:
 
         for i, (cluster, d) in enumerate(clr.items()):
             kw = _format_kw_digraph(d.get('kw', {}), base={
-                'name': 'cluster_%d' % i, 'body': {'label': '<%s>' % cluster}
+                'name': 'cluster_%d' % i, 'body': {'label': '"%s"' % cluster}
             })
             with dot.subgraph(**kw) as g:
                 for node in d['nodes']:
@@ -924,7 +950,7 @@ class SiteFolder:
             if viz:
                 viz = len(context[(self, None)].split(osp.sep)) - 1
                 viz = '/'.join(('..',) * viz + ('viz.js',))
-            dst.write(_format_output(out, viz=viz))
+            dst.write(_format_output(out, viz=viz, **kwargs))
         return {(self.view_id, None): filepath}
 
 
@@ -1051,10 +1077,7 @@ class SiteViz(SiteNode, NoView):
         self.sitemap = sitemap
 
     def render(self, context, *args, **kwargs):
-        dfl_folder = osp.join(
-            pkg_resources.resource_filename(__name__, ''), 'viz'
-        )
-        with open(osp.join(dfl_folder, 'viz.js')) as f:
+        with open(osp.join(pkg_dir, 'viz', 'viz.js')) as f:
             return f.read()
 
 
@@ -1064,9 +1087,7 @@ class SiteIndex(SiteNode):
     def __init__(self, sitemap, node_id='index'):
         super(SiteIndex, self).__init__(None, node_id, self, None, object())
         self.sitemap = sitemap
-        dfl_folder = osp.join(
-            pkg_resources.resource_filename(__name__, ''), 'index'
-        )
+        dfl_folder = osp.join(pkg_dir, 'index')
         for default_file in glob.glob(osp.join(dfl_folder, '**/*')):
             if osp.isfile(default_file):
                 self.extra_files.append(
@@ -1299,7 +1320,6 @@ class SiteIndex(SiteNode):
         )
 
     def render(self, context, *args, **kwargs):
-        pkg_dir = pkg_resources.resource_filename(__name__, '')
         fpath = osp.join(pkg_dir, 'templates', 'index.html')
 
         with open(fpath) as template:
@@ -1337,23 +1357,6 @@ def before_request(mute):
         assert request.method == method
 
 
-def _cleanup(files=None, rendered=None):
-    if files is None and rendered is None:
-        log.info('Nothing to cleanup.')
-    while files:
-        fpath = files.pop()
-        try:
-            os.remove(fpath)
-        except (FileNotFoundError, PermissionError):
-            pass
-        try:
-            os.removedirs(osp.dirname(fpath))
-        except OSError:  # The directory is not empty.
-            pass
-    rendered and rendered.clear()
-    log.info('Cleaned up generated files by the server.')
-
-
 # noinspection HttpUrlsUsage
 _repr_html = '''
 <style> .sh-box {{ width: 100%; height: 500px }} </style>
@@ -1364,12 +1367,10 @@ _repr_html = '''
 
 class ServerThread(threading.Thread):
 
-    def __init__(self, app, options):
+    def __init__(self, application, threaded=True, **kwargs):
         from werkzeug.serving import make_server
         threading.Thread.__init__(self)
-        self.srv = make_server(app=app, **options)
-        self.ctx = app.app_context()
-        self.ctx.push()
+        self.srv = make_server(app=application, threaded=threaded, **kwargs)
 
     def run(self):
         log.info('starting server')
@@ -1380,19 +1381,59 @@ class ServerThread(threading.Thread):
         self.srv.shutdown()
 
 
+class IdleContainer(threading.Thread):
+    def __init__(self, interval=1):
+        super(IdleContainer, self).__init__(daemon=True)
+        self.sites = set()
+        self.interval = interval
+        self.stop = threading.Event()
+
+    def add(self, site):
+        self.sites.add(site)
+        if not self._started.is_set():
+            self.start()
+
+    def run(self):
+        dt = self.interval + 2
+        while not self.stop.is_set():
+            time.sleep(self.interval)
+            t = time.time() - dt
+            for site in tuple(self.sites):
+                if (site.idle_timeout + site.last_activity) < t:
+                    site.shutdown()
+                    self.sites.remove(site)
+
+
+IDLES = IdleContainer()
+
+
+class ReverseProxied:
+    def __init__(self, app, script_name):
+        self.app = app
+        self.script_name = script_name
+
+    def __call__(self, environ, start_response):
+        environ['SCRIPT_NAME'] = self.script_name
+        return self.app(environ, start_response)
+
+
 class Site:
     def __init__(self, sitemap, host='localhost', port=0, delay=0.1, until=30,
-                 run_options=None, cleanup=True, **kwargs):
+                 run_options=None, idle_timeout=0, url_prefix=None, **kwargs):
         self.sitemap = sitemap
         self.kwargs = kwargs
         self.host = host
         self.port = port
         self.shutdown = lambda: False
-        self.cleanup = cleanup
         self.delay = delay
         self.until = until
+        self.idle_timeout = idle_timeout
         self._html = os.environ.get("SCHEDULA_SITE_REPR_HTML", _repr_html)
         self.run_options = {} if run_options is None else run_options
+        if url_prefix is None:
+            self.proxy = os.environ.get("SCHEDULA_FORM_URL_PREFIX", '')
+        else:
+            self.proxy = url_prefix
 
     def __repr__(self):
         s = "%s(%s, " % (self.__class__.__name__, self.sitemap)
@@ -1411,27 +1452,62 @@ class Site:
         sock.close()
         return kw
 
-    def _repr_html_(self):
-        from IPython.display import HTML
-        if not self.is_running:
-            self.run()
+    def format(self, string):
         # noinspection PyTypeChecker
         kw = combine_dicts(os.environ, {
             'id': id(self), 'host': self.host, 'port': self.port
         })
-        return HTML(self._html.format(**kw))._repr_html_()
+        return string.format(**kw)
+
+    def _repr_html_(self):
+        from IPython.display import HTML
+        if not self.is_running:
+            self.run()
+        return HTML(self.format(self._html))._repr_html_()
 
     @property
     def url(self):
         # noinspection HttpUrlsUsage
         return 'http://{}:{}'.format(self.host, self.port)
 
-    def app(self):
-        return self.sitemap.app(**self.kwargs)
+    def update_last_activity(self, response=None):
+        self.last_activity = time.time()
+        return response
 
     @staticmethod
-    def shutdown_site(shutdown, cleanup):
-        cleanup and cleanup()
+    def alive_view():
+        return 'active'
+
+    def app(self):
+        app = self.sitemap.app(**self.kwargs)
+
+        if self.idle_timeout > 0:
+            app.after_app_request(self.update_last_activity)
+            if 'alive' not in app.view_functions:
+                app.add_url_rule('/alive', 'alive', self.alive_view)
+            app.add_app_template_global(self.idle_timeout, 'alive_period')
+        from flask import Blueprint, Flask
+        if isinstance(app, Blueprint):
+            bp = app
+            app = Flask(
+                import_name=bp.import_name,
+                static_folder=bp.static_folder or 'static',
+                static_url_path=bp.static_url_path,
+                template_folder=bp.template_folder or 'templates',
+                root_path=bp.root_path
+            )
+            app.register_blueprint(bp)
+        if self.proxy:
+            app.wsgi_app = ReverseProxied(
+                app.wsgi_app, script_name=self.format(self.proxy)
+            )
+        return app
+
+    @staticmethod
+    def shutdown_site(shutdown, subsites):
+        for _, site_shutdown in subsites.values():
+            site_shutdown()
+        subsites.clear()
         shutdown()
         return True
 
@@ -1441,15 +1517,19 @@ class Site:
         memo = os.environ.get("WERKZEUG_RUN_MAIN")
         try:
             os.environ["WERKZEUG_RUN_MAIN"] = "true"
+            kw = self.get_port(**options)
             app = self.app()
-            thread = ServerThread(app, self.get_port(**options))
+            thread = ServerThread(app, **kw)
             thread.start()
             # noinspection PyArgumentList
             self.shutdown = weakref.finalize(
                 self, self.shutdown_site, thread.shutdown,
-                self.cleanup and app.cleanup
+                self.sitemap and self.sitemap.subsites or {}
             )
             self.wait_server()
+            if self.idle_timeout > 0:
+                IDLES.add(self)
+                self.update_last_activity()
             time.sleep(max(min(self.delay, 1), .1))
         finally:
             if memo is None:
@@ -1505,6 +1585,7 @@ class SiteMap(collections.OrderedDict):
         self._nodes = []
         self.foldername = ''
         self.filenames = {'index', 'index.html', 'viz.js'}
+        self.subsites = {}
 
     def __setitem__(self, key, value, *args, **kwargs):
         value.foldername = valid_filename(key, self.filenames, ext='')
@@ -1635,48 +1716,42 @@ class SiteMap(collections.OrderedDict):
     def get_directory(self, directory):
         return directory or self.directory or tempfile.mkdtemp()
 
-    def basic_app(
-            self, root_path, cleanup=None, shutdown=None, mute=True,
-            blueprint_name=None, **kwargs):
+    def basic_app(self, root_path, mute=True, blueprint_name=None, **kwargs):
         import flask
         root_path = osp.abspath(self.get_directory(root_path))
         if blueprint_name is None:
             app = flask.Flask(root_path, root_path=root_path, **kwargs)
-            app.before_request(functools.partial(before_request, mute))
-            app.mute = mute
-            app.cleanup = cleanup or _cleanup
-            app.shutdown = shutdown
         else:
             app = flask.Blueprint(
                 blueprint_name, root_path, root_path=root_path, **kwargs
             )
+        app.before_request(functools.partial(before_request, mute))
         return app
 
     def app(self, root_path=None, depth=-1, index=True, mute=True, viz_js=False,
             executor='async', blueprint_name=None, **kw):
-        generated_files, rendered = [], {}
-        cleanup = functools.partial(_cleanup, generated_files, rendered)
+        rendered = {}
         app = self.basic_app(
-            root_path, cleanup=cleanup, mute=mute,
-            blueprint_name=blueprint_name, **kw
+            root_path, mute=mute, blueprint_name=blueprint_name, **kw
         )
         context = self.rules(depth=depth, index=index, viz_js=viz_js)
 
         if context:
             rules = {v: k[0] for k, v in context.items()}
             func = functools.partial(
-                site_view, app, context, generated_files, rendered, rules,
+                site_view, context, rendered, rules,
                 next(iter(context.values())), viz=viz_js, executor=executor
             )
             app.add_url_rule('/<path:filepath>', 'default', view_func=func)
-            app.add_url_rule('/', 'default')
+            app.add_url_rule('/<string:filepath>', 'default')
+            app.add_url_rule('/', 'default', defaults={'filepath': ''})
         return app
 
-    def site(self, root_path=None, depth=-1, index=True, view=False, **kw):
+    def site(self, root_path=None, depth=-1, index=True, view=False, run=False,
+             **kw):
         site = Site(self, root_path=root_path, depth=depth, index=index, **kw)
-
+        (run or view) and site.run()
         if view:
-            site.run()
             # noinspection PyArgumentList
             self._view(site.url, format='html', quiet=True)
 
@@ -1709,7 +1784,8 @@ class SiteMap(collections.OrderedDict):
         return fpath
 
 
-def _cached_view_task(node, directory, context, rend, viz, executor):
+def _cached_view_task(
+        node, directory, context, rend, viz, executor, **render_ctx):
     n_id = node.view_id
     rend = {k: v.result() for k, v in rend.items()}
     cnt = {(n_id, e): f for (n, e), f in context.items() if n == node}
@@ -1726,7 +1802,7 @@ def _cached_view_task(node, directory, context, rend, viz, executor):
     else:
         rend = node.view(
             osp.join(directory, context[(node, None)]), context=context,
-            viz=viz, executor=executor
+            viz=viz, executor=executor, **render_ctx
         )
     return rend
 
@@ -1745,19 +1821,26 @@ def _set_rendered(results, fut, expected=None):
             )
 
 
-def cached_view(node, directory, context, rendered, viz=False,
-                executor='async'):
+def cached_view(node, directory, context, rendered, viz=False, executor='async',
+                **render_ctx):
     n_id = node.view_id
     rend = {k: v for k, v in rendered.items() if k[0] == n_id}
-    expected = {(n_id, e): f for (n, e), f in context.items() if n == node}
+    expected = {
+        (n_id, e): uncpath(osp.join(directory, f))
+        for (n, e), f in context.items()
+        if n == node
+    }
     res = {k: Future() for k in expected if k not in rendered}
-    rendered.update(res)
-    PLOT_EXECUTORS.get_executor((executor, None)).thread(
-        None, _cached_view_task, node, directory, context, rend, viz, executor
-    ).add_done_callback(functools.partial(
-        _set_rendered, res, expected=expected
-    ))
-    return combine_dicts(rend, res)
+    if res:
+        rendered.update(res)
+        PLOT_EXECUTORS.get_executor((executor, None)).thread(
+            None, _cached_view_task, node, directory, context, rend, viz,
+            executor, **render_ctx
+        ).add_done_callback(functools.partial(
+            _set_rendered, res, expected=expected
+        ))
+    rend.update(res)
+    return rend, expected
 
 
 def _compile_subs(o, n):
