@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: UTF-8 -*-
 #
-# Copyright 2015-2022, Vincenzo Arcidiacono;
+# Copyright 2015-2024, Vincenzo Arcidiacono;
 # Licensed under the EUPL (the 'Licence');
 # You may not use this work except in compliance with the Licence.
 # You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
@@ -20,6 +20,7 @@ from .cst import START
 from .gen import Token
 from .base import Base
 from .exc import DispatcherError
+from dataclasses import dataclass
 
 __author__ = 'Vincenzo Arcidiacono <vinci1it2000@gmail.com>'
 
@@ -694,10 +695,9 @@ class SubDispatch(Base):
         self.__dict__ = d
         self.__name__ = self.name
 
-    def __init__(self, dsp, outputs=None, cutoff=None, inputs_dist=None,
-                 wildcard=False, no_call=False, shrink=False,
-                 rm_unused_nds=False, output_type='all', function_id=None,
-                 output_type_kw=None):
+    def __init__(self, dsp, outputs=None, inputs_dist=None, wildcard=False,
+                 no_call=False, shrink=False, rm_unused_nds=False,
+                 output_type='all', function_id=None, output_type_kw=None):
         """
         Initializes the Sub-dispatch.
 
@@ -709,10 +709,6 @@ class SubDispatch(Base):
             Ending data nodes.
         :type outputs: list[str], iterable
 
-        :param cutoff:
-            Depth to stop the search.
-        :type cutoff: float, int, optional
-
         :param inputs_dist:
             Initial distances of input data nodes.
         :type inputs_dist: dict[str, int | float], optional
@@ -755,7 +751,6 @@ class SubDispatch(Base):
 
         self.dsp = dsp
         self.outputs = outputs
-        self.cutoff = cutoff
         self.wildcard = wildcard
         self.no_call = no_call
         self.shrink = shrink
@@ -798,17 +793,16 @@ class SubDispatch(Base):
         return memo[self]
 
     def __call__(self, *input_dicts, copy_input_dicts=False, _stopper=None,
-                 _executor=None, _sol_name=(), _verbose=False):
+                 _executor=False, _sol_name=(), _verbose=False):
 
         # Combine input dictionaries.
         i = combine_dicts(*input_dicts, copy=copy_input_dicts)
 
         # Dispatch the function calls.
         self.solution = self.dsp.dispatch(
-            i, self.outputs, self.cutoff, self.inputs_dist, self.wildcard,
-            self.no_call, self.shrink, self.rm_unused_nds,
-            stopper=_stopper, executor=_executor, sol_name=_sol_name,
-            verbose=_verbose
+            i, self.outputs, self.inputs_dist, self.wildcard, self.no_call,
+            self.shrink, self.rm_unused_nds, stopper=_stopper,
+            executor=_executor, sol_name=_sol_name, verbose=_verbose
         )
 
         return self._return(self.solution)
@@ -970,7 +964,7 @@ class MapDispatch(SubDispatch):
                  constructor=SubDispatch, constructor_kwargs=None,
                  function_id=None, func_kw=lambda *args, **data: {},
                  input_label='inputs<{}>', output_label='outputs<{}>',
-                 data_label='data<{}>', **kwargs):
+                 data_label='data<{}>', cluster_label='task<{}>', **kwargs):
         """
         Initializes the MapDispatch function.
 
@@ -1032,42 +1026,72 @@ class MapDispatch(SubDispatch):
         self.input_label = input_label
         self.output_label = output_label
         self.data_label = data_label
+        self.cluster_label = cluster_label
         self.func_kw = func_kw
 
     @staticmethod
-    def prepare_inputs(inputs, defaults, recursive_inputs=None, outputs=None):
-        if outputs and recursive_inputs:
-            data = selector(recursive_inputs, outputs or {}, allow_miss=True)
-            if isinstance(recursive_inputs, dict):
-                data = map_dict(recursive_inputs, data)
-            inputs = combine_dicts(inputs, base=data)
-        return combine_dicts(defaults, inputs)
+    def prepare_inputs(inputs, defaults):
+        inputs = [combine_dicts(defaults, d) for d in inputs]
+        return inputs if len(inputs) > 1 else inputs[0]
+
+    @staticmethod
+    def recursive_data(recursive_inputs, input_data, outputs):
+        data = selector(recursive_inputs, outputs or {}, allow_miss=True)
+        if isinstance(recursive_inputs, dict):
+            data = map_dict(recursive_inputs, data)
+        data.update(input_data)
+        return data
+
+    @staticmethod
+    def format_labels(it, label):
+        f = label.format
+        return [f(k, **v) for k, v in it]
+
+    @staticmethod
+    def format_clusters(it, label):
+        f = label.format
+        return [{'body': {
+            'label': f'"{f(k, **v)}"', 'labelloc': 'b'
+        }} for k, v in it]
 
     def _init_dsp(self, defaults, inputs, recursive_inputs=None):
-        defaults = combine_dicts(self.defaults or {}, defaults or {})
-        recursive_inputs = recursive_inputs or self.recursive_inputs
         from ..dispatcher import Dispatcher
+        defaults = combine_dicts(self.defaults or {}, defaults or {})
         self.dsp = dsp = Dispatcher(**self.kwargs)
-        func, func_kw, add_f = self.func, self.func_kw, dsp.add_function
-        pf = (defaults or recursive_inputs) and self.prepare_inputs
-        output_f, data_f = self.output_label.format, self.data_label.format
-        self.outputs, input_f = [], self.input_label.format
-        _inputs = {'defaults': defaults, 'recursive_inputs': recursive_inputs}
-        for k, v in enumerate(inputs):
-            i, o = input_f(k, **v), output_f(k, **v)
-            _inputs[i] = v
-            if pf:
-                i, keys = data_f(k, **v), [i, 'defaults']
-                if recursive_inputs and self.outputs:
-                    keys += ['recursive_inputs'] + self.outputs[-1:]
-                add_f(function=pf, inputs=keys, outputs=[i])
-            add_f(function=func, inputs=[i], outputs=[o], **func_kw(k, **v))
-            self.outputs.append(o)
-        return _inputs
+        add_data, add_func = dsp.add_data, dsp.add_func
+
+        n = len(str(len(inputs) + 1))
+        it = [(str(k).zfill(n), v) for k, v in enumerate(inputs, 1)]
+        inp = self.format_labels(it, self.input_label)
+        clt = self.format_clusters(it, self.cluster_label)
+        rl = self.format_labels(it, 'run<{}>')
+        self.outputs = out = self.format_labels(it, self.output_label)
+        add_func(self.prepare_inputs, inp, inputs=['inputs', 'defaults'])
+        recursive = recursive_inputs or self.recursive_inputs
+        if recursive:
+            func = functools.partial(self.recursive_data, recursive)
+            dat = self.format_labels(it, self.data_label)
+            fl = self.format_labels(it, 'recursive_data<{}>')
+            it = iter(zip(inp, dat, clt, fl))
+            i, d, c, fid = next(it)
+            add_data(i, clusters=c)
+            add_func(bypass, [d], inputs=[i], clusters=c)
+            for (i, d, c, fid), o, in zip(it, out[:-1]):
+                add_data(i, clusters=c)
+                add_func(func, [d], inputs=[i, o], clusters=c, function_id=fid)
+            inp = dat
+        for i, o, c, fid, (k, v) in zip(inp, out, clt, rl, enumerate(inputs)):
+            add_data(i, clusters=c)
+            kw = {'clusters': c, 'function_id': fid}
+            kw.update(self.func_kw(k, **v))
+            add_func(self.func, [o], inputs=[i], **kw)
+            add_data(o, clusters=c)
+
+        return {'inputs': inputs, 'defaults': defaults}
 
     # noinspection PyMethodOverriding
     def __call__(self, inputs, defaults=None, recursive_inputs=None,
-                 _stopper=None, _executor=None, _sol_name=(), _verbose=False):
+                 _stopper=None, _executor=False, _sol_name=(), _verbose=False):
         inputs = self._init_dsp(defaults, inputs, recursive_inputs)
         return super(MapDispatch, self).__call__(
             inputs, _stopper=_stopper, _executor=_executor, _verbose=_verbose,
@@ -1139,7 +1163,8 @@ class SubDispatchFunction(SubDispatch):
     var_keyword = 'kw'
 
     def __init__(self, dsp, function_id=None, inputs=None, outputs=None,
-                 cutoff=None, inputs_dist=None, shrink=True, wildcard=True):
+                 inputs_dist=None, shrink=True, wildcard=True, output_type=None,
+                 output_type_kw=None, first_arg_as_kw=False):
         """
         Initializes the Sub-dispatch Function.
 
@@ -1159,18 +1184,41 @@ class SubDispatchFunction(SubDispatch):
             Ending data nodes.
         :type outputs: list[str], iterable, optional
 
-        :param cutoff:
-            Depth to stop the search.
-        :type cutoff: float, int, optional
-
         :param inputs_dist:
             Initial distances of input data nodes.
         :type inputs_dist: dict[str, int | float], optional
+
+        :param shrink:
+            If True the dispatcher is shrink before the dispatch.
+        :type shrink: bool, optional
+
+        :param wildcard:
+            If True, when the data node is used as input and target in the
+            ArciDispatch algorithm, the input value will be used as input for
+            the connected functions, but not as output.
+        :type wildcard: bool, optional
+
+        :param output_type:
+            Type of function output:
+
+                + 'all': a dictionary with all dispatch outputs.
+                + 'list': a list with all outputs listed in `outputs`.
+                + 'dict': a dictionary with any outputs listed in `outputs`.
+        :type output_type: str, optional
+
+        :param output_type_kw:
+            Extra kwargs to pass to the `selector` function.
+        :type output_type_kw: dict, optional
+
+        :param first_arg_as_kw:
+            Uses the first argument of the __call__ method as `kwargs`.
+        :type output_type_kw: bool
         """
 
         if shrink:
-            dsp = dsp.shrink_dsp(inputs, outputs, cutoff=cutoff,
-                                 inputs_dist=inputs_dist, wildcard=wildcard)
+            dsp = dsp.shrink_dsp(
+                inputs, outputs, inputs_dist=inputs_dist, wildcard=wildcard
+            )
 
         if outputs:
             # Outputs not reached.
@@ -1196,16 +1244,19 @@ class SubDispatchFunction(SubDispatch):
 
         # Initialize as sub dispatch.
         super(SubDispatchFunction, self).__init__(
-            dsp, outputs, cutoff, inputs_dist, wildcard, no_call,
-            True, True, 'list'
+            dsp, outputs, inputs_dist, wildcard, no_call, True, True, 'list',
+            output_type_kw=output_type_kw
         )
-
         # Define the function to return outputs sorted.
-        if outputs is None:
+        if output_type is not None:
+            self.output_type = output_type
+        elif outputs is None:
             self.output_type = 'all'
         elif len(outputs) == 1:
             self.output_type = 'values'
 
+        self.first_arg_as_kw = first_arg_as_kw
+
     @property
     def __signature__(self):
         import inspect
@@ -1221,6 +1272,13 @@ class SubDispatchFunction(SubDispatch):
         return inspect.Signature(p, __validate_parameters__=False)
 
     def _parse_inputs(self, *args, **kw):
+        if self.first_arg_as_kw:
+            for k in sorted(args[0]):
+                if k in kw:
+                    msg = 'multiple values for argument %r'
+                    raise TypeError(msg % k) from None
+            kw.update(args[0])
+            args = args[1:]
         defaults, inputs = self.dsp.default_values, {}
         for i, k in enumerate(self.inputs or ()):
             try:
@@ -1241,8 +1299,9 @@ class SubDispatchFunction(SubDispatch):
         if self.var_keyword:
             inputs.update(kw)
         elif not all(k in inputs for k in kw):
+            k = next(k for k in sorted(kw) if k not in inputs)
             msg = 'got an unexpected keyword argument %r'
-            raise TypeError(msg % next(kw)) from None
+            raise TypeError(msg % k) from None
 
         return inputs
 
@@ -1332,8 +1391,8 @@ class SubDispatchPipe(SubDispatchFunction):
     var_keyword = None
 
     def __init__(self, dsp, function_id=None, inputs=None, outputs=None,
-                 cutoff=None, inputs_dist=None, no_domain=True, wildcard=True,
-                 shrink=True):
+                 inputs_dist=None, no_domain=True, wildcard=True, shrink=True,
+                 output_type=None, output_type_kw=None, first_arg_as_kw=False):
         """
         Initializes the Sub-dispatch Function.
 
@@ -1353,17 +1412,43 @@ class SubDispatchPipe(SubDispatchFunction):
             Ending data nodes.
         :type outputs: list[str], iterable, optional
 
-        :param cutoff:
-            Depth to stop the search.
-        :type cutoff: float, int, optional
-
         :param inputs_dist:
             Initial distances of input data nodes.
         :type inputs_dist: dict[str, int | float], optional
+
+        :param no_domain:
+            Skip the domain check.
+        :type no_domain: bool, optional
+
+        :param shrink:
+            If True the dispatcher is shrink before the dispatch.
+        :type shrink: bool, optional
+
+        :param wildcard:
+            If True, when the data node is used as input and target in the
+            ArciDispatch algorithm, the input value will be used as input for
+            the connected functions, but not as output.
+        :type wildcard: bool, optional
+
+        :param output_type:
+            Type of function output:
+
+                + 'all': a dictionary with all dispatch outputs.
+                + 'list': a list with all outputs listed in `outputs`.
+                + 'dict': a dictionary with any outputs listed in `outputs`.
+        :type output_type: str, optional
+
+        :param output_type_kw:
+            Extra kwargs to pass to the `selector` function.
+        :type output_type_kw: dict, optional
+
+        :param first_arg_as_kw:
+            Converts first argument of the __call__ method as `kwargs`.
+        :type output_type_kw: bool
         """
 
         self.solution = sol = dsp.solution.__class__(
-            dsp, inputs, outputs, wildcard, cutoff, inputs_dist, True, True,
+            dsp, inputs, outputs, wildcard, inputs_dist, True, True,
             no_domain=no_domain
         )
         sol._run()
@@ -1374,8 +1459,9 @@ class SubDispatchPipe(SubDispatchFunction):
             dsp = dsp._get_dsp_from_bfs(o, bfs_graphs=bfs)
 
         super(SubDispatchPipe, self).__init__(
-            dsp, function_id, inputs, outputs=outputs, cutoff=cutoff,
-            inputs_dist=inputs_dist, shrink=False, wildcard=wildcard
+            dsp, function_id, inputs, outputs=outputs, inputs_dist=inputs_dist,
+            shrink=False, wildcard=wildcard, output_type=output_type,
+            output_type_kw=output_type_kw, first_arg_as_kw=first_arg_as_kw
         )
         self._reset_sol()
         self.pipe = self._set_pipe()
@@ -1442,7 +1528,7 @@ class SubDispatchPipe(SubDispatchFunction):
                 break
 
             for n, vw_d in nxt_dsp:
-                s._set_sub_dsp_node_input(v, n, [], s.check_cutoff, False, vw_d)
+                s._set_sub_dsp_node_input(v, n, [], False, vw_d)
             s._see_remote_link_node(v)
 
         # Return outputs sorted.
@@ -1618,9 +1704,17 @@ def add_function(dsp, inputs_kwargs=False, inputs_defaults=False, **kw):
     return decorator
 
 
-class inf(collections.namedtuple('_inf', ['inf', 'num'])):
+@dataclass(repr=False, frozen=True, eq=False)
+class inf:
     """Class to model infinite numbers for workflow distance."""
 
+    _inf: float = 0
+    _num: float = 0
+
+    def __iter__(self):
+        yield self._inf
+        yield self._num
+
     @staticmethod
     def format(val):
         if not isinstance(val, tuple):
@@ -1628,14 +1722,14 @@ class inf(collections.namedtuple('_inf', ['inf', 'num'])):
         return inf(*val)
 
     def __repr__(self):
-        if self[0] == 0:
-            return str(self[1])
+        if self._inf == 0:
+            return str(self._num)
         return 'inf(inf={}, num={})'.format(*self)
 
     def __add__(self, other):
         if isinstance(other, self.__class__):
-            return inf(self[0] + other[0], self[1] + other[1])
-        return inf(self[0], self[1] + other)
+            return inf(self._inf + other._inf, self._num + other._num)
+        return inf(self._inf, self._num + other)
 
     def __sub__(self, other):
         other = isinstance(other, self.__class__) and other or (0, other)
